// TEd.java - A Tiny Editor Command for Slush.
//
// Copyright (C) 1999,2000,2001  Smart Software Consulting
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
// Smart Software Consulting
// 1688 Silverwood Court
// Danville, CA  94526-3079
// USA
//
// http://www.smartsc.com
//

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.PrintStream;
import java.util.Hashtable;
import java.util.StringTokenizer;
import java.util.Vector;

import com.dalsemi.shell.server.SystemInputStream;
import com.dalsemi.shell.server.SystemPrintStream;
import com.dalsemi.slush.CommandInterpreter;
import com.dalsemi.slush.Slush;
import com.dalsemi.slush.command.SlushCommand;

public class TEd
{
	public static class Command
	implements SlushCommand
	{
		public void execute(
			SystemInputStream in,
			SystemPrintStream out,
			SystemPrintStream err,
			String[] args,
			Hashtable env)
		{
			try
			{
				new TEd( in, out, err, args, env);
			}
			catch( Exception e)
			{
				err.println( "TEdCommand: Got exception: " + e);
			}
		}

		public String getUsageString()
		{
			return USAGE;
		}
	}

	TEd(
		SystemInputStream in,
		SystemPrintStream out,
		SystemPrintStream err,
		String[] args,
		Hashtable env)
	{
		if( args.length < 1)
		{
			err.println( USAGE);
			return;
		}

		out.println( PROGRAM_INFO);
		out.println( COPYRIGHT);
		out.println();

		this.in  = in;
		this.out = out;
		this.err = err;
		filename = args[0];

		try
		{
			modified = false;
			open();
			run();
		}
		catch( Exception e)
		{
			err.println( "TEd: Got exception: " + e);
		}
	}

	public void open()
	throws IOException
	{
		saveChanges();

		clearBuffer();

		File f = new File( filename);
		filename = f.getAbsolutePath();

		if( f.isDirectory())
			throw new IOException( "Can't edit directory!");

		if( f.exists())
		{
			readFile( f);

			modified = false;
			readOnly = false;
			if( !f.canWrite())
			{
				out.println(
					"Opening " + filename + " in read-only mode.");
				readOnly = true;
			}
		}
		else
		{
			modified = true;

			out.print( "New file '");
			out.print( filename);
			out.println( "'");
		}
	}

	public void readFile( File f)
	throws IOException
	{
		BufferedReader br = null;
		int oldCurrentLine = currentLine;
		try
		{
			br =
				new BufferedReader(
					new InputStreamReader(
						new FileInputStream( f)));

			for( String s = br.readLine();
			     s != null;
			     s = br.readLine())
			{
				insertLine( s);
			}

			out.print( "Read ");
			out.print( currentLine - oldCurrentLine);
			out.print( " line");
			if( lines.size() != 1) out.print( "s");
			out.print( " from '");
			out.print( f.getAbsolutePath());
			out.println( "'");
		}
		catch( IOException ioe)
		{
			// Remove inserted lines
			while( currentLine > oldCurrentLine)
			{
				lines.removeElementAt( currentLine--);
			}
			// Rethrow IOException
			throw ioe;
		}
		finally
		{
			if( br != null) br.close();
		}
	}

	public void run()
	{
		String cmd;

		quit = false;
		while( !quit)
		{
			out.print( PROMPT);

			cmd = in.readLine();

			try { parseCommand( cmd); }
			catch( Exception e)
			{
				out.println( "Invalid command.");
			}
		}
	}

	public void parseCommand( String cmd)
	throws IllegalArgumentException
	{
		if( cmd == null || cmd.length() == 0)
		{
			return;
		}

		String name = null;

		// Index of first whitespace
		int wsIdx = 1;
		switch( cmd.charAt( 0))
		{
			case 'a':
				if( cmd.startsWith( APPEND))
				{
					wsIdx = APPEND.length();
				}

				if( cmd.charAt( wsIdx) != ' ' )
				{
					throw new IllegalArgumentException();
				}

				append( cmd.substring( wsIdx+1));
				break;

			case 'c':
				// Must be c/for/bar or cl buff
				if( cmd.length() == 1)
				{
					throw new IllegalArgumentException();
				}
				if( cmd.startsWith( CHANGE))
				{
					wsIdx = CHANGE.length()+1;
				}
				if( cmd.charAt( 1) != 'l')
				{
					change( cmd.substring( wsIdx));
					break;
				}
				// else, clear buffer
				clearBuffer();
				break;

			case 'd':
				if( !cmd.startsWith( DELETE))
				{
					throw new IllegalArgumentException();
				}

				wsIdx = DELETE.length();

				if( wsIdx == cmd.length())
				{
					delete();
				}
				else
				{
					delete( cmd.substring( wsIdx+1));
				}
				break;

			case 'i':
				if( cmd.startsWith( INSERT))
				{
					wsIdx = INSERT.length();
				}

				if( wsIdx == cmd.length())
				{
					insert();
				}
				else if( cmd.charAt( wsIdx) != ' ' )
				{
					throw new IllegalArgumentException();
				}
				else
				{
					insertLine( cmd.substring( wsIdx+1));
					listCurrentLine();
				}
				break;

			case 'l':
				if( cmd.startsWith( LIST))
				{
					wsIdx = LIST.length();
				}

				list( cmd.substring( wsIdx));
				break;

			case 'o':
				if( cmd.startsWith( OPEN))
				{
					wsIdx = OPEN.length();
				}

				if( cmd.charAt( wsIdx) != ' ' )
				{
					throw new IllegalArgumentException();
				}

				name = cmd.substring( wsIdx+1).trim();
				if( name.startsWith( "\""))
				{
					name = name.substring( 1, name.length() - 1);
				}

				filename = name;
				try { open(); }
				catch( IOException ioe)
				{
					out.print( "Error opening file '");
					out.print( name);
					out.print( "' (");
					out.print( ioe.getMessage());
					out.println( ")");
					quit = true;
				}
				break;

			case 'q':
				saveChanges();
				quit = true;
				break;

			case 'r':
				if( cmd.startsWith( READ))
				{
					wsIdx = READ.length();
				}

				if( cmd.charAt( wsIdx) != ' ' )
				{
					throw new IllegalArgumentException();
				}

				name = cmd.substring( wsIdx+1).trim();
				if( name.startsWith( "\""))
				{
					name = name.substring( 1, name.length() - 1);
				}
				if( name.length() == 0)
				{
					out.println( "No filename given.");
					break;
				}

				File f = new File( filename);
				try
				{
					readFile( f);
					modified = true;
				}
				catch( IOException ioe)
				{
					out.print( "Error reading file '");
					out.print( f.getAbsolutePath());
					out.print( "' (");
					out.print( ioe.getMessage());
					out.println( ")");
				}
				break;

			case 'w':
				if( cmd.startsWith( WRITE))
				{
					wsIdx = WRITE.length();
				}

				name = filename;
				if( wsIdx == cmd.length() && readOnly)
				{
					out.print( "File '");
					out.print( filename);
					out.println( "' is read-only.");
					break;
				}
				else if( wsIdx < cmd.length())
				{
					if( cmd.charAt( wsIdx) != ' ' )
					{
						throw new IllegalArgumentException();
					}

					name = cmd.substring( wsIdx+1).trim();
					if( name.startsWith( "\""))
					{
						name = name.substring( 1, name.length() - 1);
					}
				}
				write( name);
				break;

			case '0':
			case '1':
			case '2':
			case '3':
			case '4':
			case '5':
			case '6':
			case '7':
			case '8':
			case '9':
			case '#':
				StringTokenizer st = new StringTokenizer( cmd);
				String sNum = st.nextToken();
				int num = Integer.decode( sNum).intValue()-1;
				if( st.hasMoreTokens())
				{
					setCurrentLine( num);
					setCurrentLine( st.nextToken());
				}
				list( num, num);
				break;

			default:
				throw new IllegalArgumentException();
		}
	}

	// Append related methods
	protected void append( String data)
	throws IllegalArgumentException
	{
		if( data.length() > 0)
		{
			getCurrentLine().append( data);
			modified = true;
		}
		listCurrentLine();
	}

	// List related methods
	protected void list( String data)
	throws IllegalArgumentException
	{
		data.trim();
		StringTokenizer st = new StringTokenizer( data);
		if( !st.hasMoreTokens())
		{
			list( 0, lines.size()-1);
			return;
		}

		String sStart = st.nextToken();
		int start = currentLine;
		if( !STAR.equals( sStart))
		{
			start = Integer.decode( sStart).intValue() - 1;
		}

		if( !st.hasMoreTokens())
		{
			list( start, start);
			return;
		}

		String sEnd = st.nextToken();
		int end = currentLine;
		if( !STAR.equals( sEnd))
		{
			end = Integer.decode( sEnd).intValue() - 1;
		}

		if( !st.hasMoreTokens())
		{
			list( start, end);
			return;
		}

		throw new IllegalArgumentException();
	}

	protected void list( int start, int end)
	{
		if( start < 0) start = 0;
		if( end < 0)   end   = 0;

		if( start >= lines.size()) start = lines.size()-1;
		if( end   >= lines.size()) end   = lines.size()-1;

		// If buffer empty
		if( start < 0)
		{
			return;
		}

		currentLine = end;

		if( start > end)
		{
			int tmp = start;
			start = end;
			end = tmp;
		}

		for( int line = start; line <= end; ++line)
		{
			out.print( line+1);
			out.print( ": ");
			if( line == currentLine)
				out.print( "*");
			else
				out.print( " ");
			out.println( (StringBuffer)lines.elementAt( line));
		}

		out.println();
	}

	protected void listCurrentLine()
	{
		list( currentLine, currentLine);
	}

	// Delete related methods
	protected void delete( String data)
	throws IllegalArgumentException
	{
		data.trim();
		StringTokenizer st = new StringTokenizer( data);
		if( !st.hasMoreTokens())
		{
			delete( currentLine, currentLine);
			return;
		}

		String sStart = st.nextToken();
		int start = currentLine;
		if( !STAR.equals( sStart))
		{
			start = Integer.decode( sStart).intValue() - 1;
		}

		if( !st.hasMoreTokens())
		{
			delete( start, start);
			return;
		}

		String sEnd = st.nextToken();
		int end = currentLine;
		if( !STAR.equals( sEnd))
		{
			end = Integer.decode( sEnd).intValue() - 1;
		}

		if( !st.hasMoreTokens())
		{
			delete( start, end);
			return;
		}

		throw new IllegalArgumentException();
	}

	protected void delete()
	{
		delete( currentLine, currentLine);
	}

	protected void delete( int start, int end)
	{
		if( start < 0) start = 0;
		if( end < 0)   end   = 0;

		if( start >= lines.size()) start = lines.size()-1;
		if( end   >= lines.size()) end   = lines.size()-1;

		// If buffer empty
		if( start < 0)
		{
			return;
		}

		if( start > end)
		{
			int tmp = start;
			start = end;
			end = tmp;
		}

		for( int line = end; line >= start; --line)
		{
			lines.removeElementAt( line);
		}

		if( end >= start)
			modified = true;

		currentLine = start - 1;
		if( currentLine < 0 && lines.size() > 0)
			currentLine = 0;
	}

	protected void deleteCurrentLine()
	{
		delete( currentLine, currentLine);
	}

	// Clear Buffer
	protected void clearBuffer()
	{
		if( currentLine != -1)
		{
			lines.removeAllElements();
			currentLine = -1;
			modified = true;
		}
	}

	// Change related methods
	protected void change( String cmd)
	{
		String oldText = null;
		String newText = null;
		String opts    = null;
		boolean global = false;

		StringTokenizer st = new StringTokenizer( cmd, cmd.substring( 0, 1));
		if( st.hasMoreTokens())
			oldText = st.nextToken();
		if( st.hasMoreTokens())
			newText = st.nextToken();
		if( st.hasMoreTokens())
		{
			opts = st.nextToken();
			if( opts.indexOf( 'g') != -1)
				global = true;
		}

		boolean found = false;
		do
		{
			found = change( oldText, newText);
		} while( global && found);

		listCurrentLine();
	}

	protected boolean change( String oldText, String newText)
	{
		if( newText == null) newText = "";
		String s = getCurrentLine().toString();
		int idx = s.indexOf( oldText);
		if( idx == -1)
			return false;
		StringBuffer sb = new StringBuffer(
			s.length() - oldText.length() + newText.length() );
		sb.append( s.substring( 0, idx));
		sb.append( newText);
		if( idx + oldText.length() < s.length())
			sb.append( s.substring( idx + oldText.length()));
		setCurrentLine( sb);
		modified = true;
		return true;
	}

	// Insert related methods
	protected void insert()
	{
		String line;
		// Loop until break
		while( true)
		{
			out.print( currentLine + 2);
			out.print( "i: ");

			line = in.readLine();

			if( ".".equals( line))
				break;
			if( line.startsWith( ".."))
				insertLine( line.substring(1));
			else
				insertLine( line);
			modified = true;
		}
	}

	protected void write( String name)
	{
		if( name.length() == 0)
		{
			out.println( "No filename given.");
			return;
		}

		File f = new File( name);
		if( f.isDirectory())
		{
			out.println( "Cannot write directories.");
			return;
		}

		// If exists but is not writeable
		if( f.exists() && !f.canWrite())
		{
			out.print( "Cannot write to '");
			out.print( f.getAbsolutePath());
			out.println( "'");
			return;
		}

		BufferedWriter bw = null;
		try
		{
			bw =
				new BufferedWriter(
					new OutputStreamWriter(
						new FileOutputStream( name)));

			for( int i = 0; i < lines.size(); ++i)
			{
				bw.write( lines.elementAt( i).toString());
				bw.newLine();
			}

			filename = name;
			modified = false;

			out.print( "Wrote ");
			out.print( lines.size());
			out.print( " line");
			if( lines.size() != 1) out.print( "s");
			out.print( " to '");
			out.print( f.getAbsolutePath());
			out.println( "'");
		}
		catch( IOException ioe)
		{
			out.println( "Error writing to '");
			out.print( f.getAbsolutePath());
			out.println( "'");
		}
		finally
		{
			if( bw != null)
			{
				try { bw.close(); }
				catch( IOException ioe) {}
			}
		}
	}

	protected void saveChanges()
	{
		if( modified)
		{
			String cmd = "";

			while( cmd.length() == 0)
			{
				out.print( "Save changes to '");
				out.print( filename);
				out.print( "'? [y/n] ");

				cmd = in.readLine().trim();
				if( cmd.length() > 0)
				{
					char ch = cmd.charAt( 0);
					if( ch == 'Y' || ch == 'y')
					{
						write( filename);
					}
				}
			}
		}
	}

	protected SystemInputStream in  = null;
	protected SystemPrintStream out = null;
	protected SystemPrintStream err = null;

	protected boolean modified = false;
	protected boolean quit = false;

	protected String filename = null;

	protected boolean readOnly = false;
	protected int currentLine = -1;
	protected Vector lines = new Vector();
	protected StringBuffer getCurrentLine()
	{
		return (StringBuffer)lines.elementAt( currentLine);
	}
	protected void setCurrentLine( StringBuffer sb)
	{
		if( currentLine < 0)
			insertLine( sb);
		else
			lines.setElementAt( sb, currentLine);
		modified = true;
	}
	protected void setCurrentLine( String s)
	{
		setCurrentLine( new StringBuffer( s));
	}
	protected void setCurrentLine( int i)
	{
		if( lines.size() > 0)
		{
			currentLine = i;
			if( currentLine < 0)
				currentLine = 0;
			if( currentLine >= lines.size())
				currentLine = lines.size() - 1;
		}
	}
	protected void insertLine( StringBuffer sb)
	{
		lines.insertElementAt( sb, ++currentLine);
		modified = true;
	}
	protected void insertLine( String s)
	{
		insertLine( new StringBuffer( s));
	}

	// This main method is a subtitute for Slush.main.  It is needed
	// because Slush.initializeShellCommands() is private instead of
	// protected. :-(  So we must add the TEd.Command instance to the
	// CommandInterpreter *after* calling Sliush.main().
	public static void main( String[] args)
	{
		Slush.main( args);
		CommandInterpreter.addCommand( "ted", new TEd.Command());
	}

	public static final String PROGRAM_INFO = "TEd Version 0.8";
	public static final String COPYRIGHT =
		"Copyright (C) 1999,2000,2001 Smart Software Consulting.";
	public static final String HELP_1 = "ted FILENAME";
	public static final String HELP_2 = "Edits FILENAME using TEd.";
	public static final String HELP_3 =
		"For more information, see 'http://www.smartsc.com/tini/TEd'.";

	public static final String USAGE;
	static
	{
		String lineSep = System.getProperty( "line.separator");
		StringBuffer sb = new StringBuffer( 200);
		sb.append( PROGRAM_INFO); sb.append( lineSep);
		sb.append( COPYRIGHT); sb.append( lineSep); sb.append( lineSep);
		sb.append( HELP_1); sb.append( lineSep); sb.append( lineSep);
		sb.append( HELP_2); sb.append( lineSep);
		sb.append( HELP_3);
		USAGE = sb.toString();
	}

	public static final String PROMPT = "TEd> ";

	public static final String APPEND = "append";
	public static final String CHANGE = "change";
	public static final String DELETE = "del";
	public static final String INSERT = "insert";
	public static final String LIST   = "list";
	public static final String OPEN   = "open";
	public static final String QUIT   = "quit";
	public static final String READ   = "read";
	public static final String STAR   = "*";
	public static final String WRITE  = "write";
}